from datetime import datetime

import alembic.command
from alembic.config import Config as AlembicConfig
from sqlalchemy import Column, DateTime, Integer, String, create_engine, select
from sqlalchemy.orm import declarative_base, sessionmaker

from ..config import Database as DatabaseConfig

Base = declarative_base()


class Execution(Base):
    __tablename__ = "executions"

    id = Column(Integer, primary_key=True)
    executed_at = Column(DateTime, nullable=False)


class Run(Base):
    __tablename__ = "runs"

    id = Column(Integer, primary_key=True)
    checksum = Column(String(length=32), nullable=False)
    name = Column(String(length=255), nullable=False)


class Database:
    def __init__(self, engine):
        self.session: sessionmaker = sessionmaker(engine, expire_on_commit=False)

    def get_last_execution(self) -> Execution:
        stmt = select(Execution).order_by(Execution.executed_at.desc())
        with self.session() as session:
            execution = session.scalars(stmt).first()
            if execution:
                return execution
            else:
                ex = Execution()
                ex.executed_at = datetime.fromtimestamp(0.0)
                return ex

    def get_or_create_task(self, name: str, checksum: str = "") -> Run:
        stmt = select(Run).where(Run.name == name)
        with self.session() as session, session.begin():
            run = session.scalars(stmt).first()
            if run:
                return run

            run = Run()
            run.checksum = checksum
            run.name = name
            session.add(run)
            return run

    def update_task(self, name: str, checksum: str) -> None:
        stmt = select(Run).where(Run.name == name)
        with self.session() as session, session.begin():
            run = session.scalars(stmt).first()
            run.checksum = checksum
            session.add(run)
            return

    def save_execution(self, execution: Execution):
        with self.session() as session, session.begin():
            session.add(execution)
            session.commit()


def new_database(cfg: DatabaseConfig) -> Database:
    engine = create_engine(cfg.connection)
    if cfg.migrate is True:
        with engine.begin() as connection:
            alembic_config = AlembicConfig()
            alembic_config.set_main_option(
                "script_location", "rcmt:database:migrations"
            )
            alembic_config.attributes["connection"] = connection
            alembic.command.upgrade(config=alembic_config, revision="head")

    return Database(engine)



import datetime
import fnmatch
import io
import os.path
from typing import Any, Generator, Optional, TextIO, Union
from urllib.parse import urlparse

import gitlab
import structlog
from gitlab import GitlabGetError
from gitlab.base import RESTObjectList
from gitlab.v4.objects import CurrentUser
from gitlab.v4.objects import Project as GitlabProject
from gitlab.v4.objects.merge_requests import ProjectMergeRequest as GitlabMergeRequest

from ..log import SECRET_MASKER
from .source import Base, PullRequest, Repository, add_credentials_to_url

log: structlog.stdlib.BoundLogger = structlog.get_logger(source="gitlab")


class GitlabRepository(Repository):
    def __init__(self, project: GitlabProject, token: str, url: str):
        self._project = project
        self.token = token
        self.url = url

    @property
    def base_branch(self) -> str:
        return self._project.default_branch

    def can_merge_pull_request(self, identifier: GitlabMergeRequest) -> bool:
        return True

    @property
    def clone_url(self) -> str:
        # Value of username does not matter to GitLab.
        return add_credentials_to_url(
            url=self._project.http_url_to_repo, password=self.token, username="rcmt"
        )

    def close_pull_request(self, message: str, pr: GitlabMergeRequest) -> None:
        pr.notes.create({"body": message})
        pr.state_event = "close"
        pr.save()

    def create_pull_request(self, branch: str, pr: PullRequest) -> None:
        log.debug(
            "Creating merge request", base=self.base_branch, head=branch, repo=str(self)
        )
        self._project.mergerequests.create(
            {
                "description": pr.body,
                "source_branch": branch,
                "target_branch": self.base_branch,
                "title": pr.title,
            }
        )

    def delete_branch(self, identifier: GitlabMergeRequest) -> None:
        if identifier.should_remove_source_branch is not True:
            self._project.branches.get(id=identifier.source_branch, lazy=True).delete()

    def find_pull_request(self, branch: str) -> Union[Any, None]:
        log.debug("Listing merge requests", repo=str(self))
        mrs = self._project.mergerequests.list(
            all=False, state="all", source_branch=branch
        )
        if len(mrs) == 0:
            return None

        if isinstance(mrs, list):
            return mrs[0]

        if isinstance(mrs, RESTObjectList):
            return mrs.next()

        raise RuntimeError(
            f"GitLab API returned an unexpected object {mrs.__class__.__name__}"
        )

    def get_file(self, path: str) -> TextIO:
        try:
            content = self._project.files.get(
                file_path=path, ref=self.base_branch
            ).decode()
            if content is None:
                log.warning("Decoded content is None", repo=str(self), file=path)
                raise FileNotFoundError("decoded content is None")

            return io.StringIO(content.decode("utf-8"))
        except GitlabGetError:
            raise FileNotFoundError("file does not exist in repository")

    def has_file(self, path: str) -> bool:
        directory = os.path.dirname(path)
        file = os.path.basename(path)
        try:
            tree = self._project.repository_tree(path=directory, iterator=True)
            for entry in tree:
                if entry["type"] != "blob":
                    continue

                if fnmatch.fnmatch(entry["path"], file):
                    return True
        except gitlab.GitlabGetError as e:
            if e.response_code == 404:
                log.warning("Tree not found - empty repository?", repo=str(self))
                return False
            else:
                raise e

        return False

    def has_successful_pr_build(self, identifier: GitlabMergeRequest) -> bool:
        if identifier.approvals.get().approved is False:
            log.debug("Approvals missing", repo=str(self), id=identifier.get_id())
            return False

        failed = False
        for commit_status in self._project.commits.get(
            id=identifier.sha, lazy=True
        ).statuses.list(iterator=True):
            if commit_status.allow_failure is True:
                continue

            if commit_status.status != "success":
                log.debug(
                    "MR check failed",
                    repo=str(self),
                    id=identifier.get_id(),
                    name=commit_status.name,
                    status=commit_status.status,
                )
                failed = True

        if failed is False:
            log.debug(
                "All MR checks successful", repo=str(self), id=identifier.get_id()
            )
            return True

        return False

    def is_pr_closed(self, mr: GitlabMergeRequest) -> bool:
        return mr.state == "closed" or mr.state == "locked"

    def is_pr_merged(self, mr: GitlabMergeRequest) -> bool:
        return mr.state == "merged"

    def is_pr_open(self, mr: GitlabMergeRequest) -> bool:
        return mr.state == "opened"

    def merge_pull_request(self, identifier: GitlabMergeRequest):
        log.debug("Merging merge request", repo=str(self), id=identifier.get_id())
        identifier.merge()

    @property
    def name(self) -> str:
        return self._project.path

    def pr_created_at(self, pr: GitlabMergeRequest) -> datetime.datetime:
        corrected: str = pr.created_at.replace("Z", "")
        return datetime.datetime.fromisoformat(corrected)

    @property
    def project(self) -> str:
        return self._project.namespace["full_path"]

    @property
    def source(self) -> str:
        return self.url

    def update_pull_request(self, pr: GitlabMergeRequest, pr_data: PullRequest) -> None:
        needs_update = False
        if pr.title != pr_data.title:
            pr.title = pr_data.title
            needs_update = True

        if pr.description != pr_data.body:
            pr.description = pr_data.body
            needs_update = True

        if needs_update is True:
            log.debug("Updating MR data", id=pr.get_id(), repo=str(self))
            pr.save()


class Gitlab(Base):
    def __init__(self, url: str, private_token: str):
        self.client = gitlab.Gitlab(url, private_token=private_token)
        self.url = urlparse(url).netloc

        SECRET_MASKER.add_secret(private_token)

    def create_from_name(self, name: str) -> Optional[Repository]:
        if name.startswith(self.url) is False:
            return None

        name_without_host = name.replace(f"{self.url}/", "")
        try:
            p = self.client.projects.get(id=name_without_host)
        except gitlab.GitlabGetError as e:
            log.debug("Unable to get project", name=name, status_code=e.response_code)
            return None

        token = self.client.private_token or ""
        return GitlabRepository(project=p, token=token, url=self.url)

    def list_repositories_with_open_pull_requests(
        self,
    ) -> Generator[Repository, None, None]:
        if self.client.user is None:
            self.client.auth()

        user: Optional[CurrentUser] = self.client.user
        if user is None:
            log.error("unable to authenticate user while listing open pull requests")
            return

        merge_requests = self.client.mergerequests.list(
            author_id=user.attributes["id"], state="opened"
        )
        seen_project_ids: list[int] = []
        for mr in merge_requests:
            if mr.project_id in seen_project_ids:
                continue

            seen_project_ids.append(mr.project_id)
            project = self.client.projects.get(id=mr.project_id)
            repository = GitlabRepository(
                project=project, token=self.client.private_token, url=self.url  # type: ignore
            )
            yield repository

    def list_repositories(self, since: datetime.datetime) -> list[Repository]:
        log.debug("start fetching repositories")
        projects = self.client.projects.list(
            all=True, archived=False, min_access_level=30, last_activity_after=since
        )
        repositories: list[Repository] = []
        for p in projects:
            repositories.append(GitlabRepository(project=p, token=self.client.private_token, url=self.url))  # type: ignore

        log.debug("finished fetching repositories")
        return repositories



import os.path
from typing import Any, Mapping, Union

import git
import structlog
from git.exc import GitCommandError

from rcmt import source

log = structlog.get_logger(package="git")


class Git:
    def __init__(
        self,
        branch_name: str,
        clone_opts: Mapping[str, Any],
        data_dir: str,
        user_name: str,
        user_email: str,
    ):
        self.branch_name = branch_name
        self.clone_opts = clone_opts
        self.data_dir = data_dir
        self.user_email = user_email
        self.user_name = user_name

    def checkout_dir(self, repo: source.Repository) -> str:
        return os.path.join(self.data_dir, repo.source, repo.project, repo.name)

    def commit_changes(self, repo_dir: str, msg: str):
        git_repo = git.Repo(path=repo_dir)
        git_repo.git.add(all=True)
        git_repo.index.commit(msg)

    @staticmethod
    def has_changes(repo_dir: str) -> bool:
        git_repo = git.Repo(path=repo_dir)
        return len(git_repo.index.diff(None)) > 0 or len(git_repo.untracked_files) > 0

    def has_changes_base(self, base_branch: str, repo_dir: str) -> bool:
        git_repo = git.Repo(path=repo_dir)
        return len(git_repo.index.diff(f"origin/{base_branch}")) > 0

    def needs_push(self, repo_dir: str) -> bool:
        git_repo = git.Repo(path=repo_dir)
        try:
            return len(git_repo.index.diff(f"origin/{self.branch_name}")) > 0
        except git.BadName:
            return True

    def prepare(self, repo: source.Repository) -> str:
        checkout_dir = self.checkout_dir(repo)
        if os.path.exists(checkout_dir) is False:
            log.debug("Cloning repository", url=repo.clone_url, repo=str(repo))
            os.makedirs(checkout_dir)
            git_repo = git.Repo.clone_from(
                repo.clone_url, checkout_dir, **self.clone_opts
            )
        else:
            git_repo = git.Repo(path=checkout_dir)
            self.restore(git_repo)

        if self.validate_branch_name(git_repo) is False:
            raise RuntimeError(f"Branch name '{self.branch_name}' is not valid")

        git_repo.config_writer().set_value("user", "email", self.user_email).release()
        git_repo.config_writer().set_value("user", "name", self.user_name).release()
        log.debug("Checking out base branch", branch=repo.base_branch, repo=str(repo))
        git_repo.heads[repo.base_branch].checkout()
        hash_before_pull = str(git_repo.head.commit)
        log.debug(
            "Pulling changes into base branch", branch=repo.base_branch, repo=str(repo)
        )
        git_repo.remotes["origin"].pull()
        git_repo.git.remote("prune", "origin")
        hash_after_pull = str(git_repo.head.commit)
        has_base_branch_update = hash_before_pull != hash_after_pull
        if has_base_branch_update is True:
            log.debug(
                "Base branch contains new commits",
                base_branch=repo.base_branch,
                repo=str(repo),
            )

        exists_local = branch_exists_local(self.branch_name, git_repo)
        remote_branch = get_remote_branch(self.branch_name, git_repo)
        if exists_local is False:
            log.debug("Creating branch", branch=self.branch_name, repo=str(repo))
            if remote_branch is None:
                git_repo.create_head(self.branch_name)
            else:
                git_repo.create_head(self.branch_name, remote_branch)

        has_conflict = False
        if remote_branch is not None:
            try:
                # Try to merge. Errors if there is a merge conflict.
                git_repo.git.merge(self.branch_name, no_ff=True, no_commit=True)
            except GitCommandError as e:
                # Exit codes "1" or "2" indicate that a merge is not successful
                if e.status != 1 and e.status != 2:
                    raise e

                log.debug(
                    "Merge conflict with base branch",
                    base_branch=repo.base_branch,
                    repo=str(repo),
                )
                has_conflict = True

            try:
                # Abort the merge to not leave the branch in a conflicted state
                git_repo.git.merge(abort=True)
            except GitCommandError as e:
                # "128" is the exit code of the git command if no abort was needed
                if e.status != 128:
                    raise e

        log.debug("Checking out work branch", branch=self.branch_name, repo=str(repo))
        git_repo.heads[self.branch_name].checkout()
        if has_conflict is True or has_base_branch_update is True:
            merge_base = git_repo.git.merge_base(repo.base_branch, self.branch_name)
            log.debug(
                "Resetting to merge base", branch=self.branch_name, repo=str(repo)
            )
            git_repo.git.reset(merge_base, hard=True)
            log.debug(
                "Rebasing onto work branch", branch=self.branch_name, repo=str(repo)
            )
            git_repo.git.rebase(repo.base_branch)

        return checkout_dir

    def push(self, repo_dir):
        git_repo = git.Repo(path=repo_dir)
        git_repo.git.push("origin", self.branch_name, force=True, set_upstream=True)

    def restore(self, repo: git.Repo) -> None:
        repo.git.restore(".")

    def validate_branch_name(self, repo: git.Repo) -> bool:
        try:
            repo.git.check_ref_format(f"refs/heads/{self.branch_name}")
            return True
        except GitCommandError:
            return False


def branch_exists_local(name: str, repo: git.Repo) -> bool:
    for b in repo.heads:
        if b.name == name:
            return True

    return False


def get_remote_branch(name: str, repo: git.Repo) -> Union[git.RemoteReference, None]:
    remote = repo.remote()
    for r in remote.refs:
        if r.name == f"{remote.name}/{name}":
            return r

    return None



import importlib.metadata

from .rcmt import execute, options_from_config
from .task import Run, Task
from .validate import validate
from .verify import execute as execute_verify

__all__ = [
    "Run",
    "Task",
    "execute",
    "execute_verify",
    "options_from_config",
    "validate",
]

try:
    __version__ = importlib.metadata.version(__name__)
except Exception:
    __version__ = "develop"



"""Add model "Run"

Revision ID: 1ebf5b1ca7fe
Revises: b3117b16eb76
Create Date: 2022-12-17 20:49:10.111787

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "1ebf5b1ca7fe"
down_revision = "b3117b16eb76"
branch_labels = None
depends_on = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "runs",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("checksum", sa.String(length=32), nullable=False),
        sa.Column("name", sa.String(length=255), nullable=False),
        sa.PrimaryKeyConstraint("id"),
    )
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table("runs")
    # ### end Alembic commands ###



import datetime
from enum import Enum
from typing import Optional

import structlog

from . import config, database, encoding, git, source, task

log: structlog.stdlib.BoundLogger = structlog.get_logger()


class Options:
    def __init__(self, cfg: config.Config):
        self.config = cfg
        self.encoding_registry: encoding.Registry = encoding.Registry()
        self.task_paths: list[str] = []
        self.sources: dict[str, source.Base] = {}


def can_merge_after(
    created_at: datetime.datetime, delay: Optional[datetime.timedelta]
) -> bool:
    if delay is None:
        return True

    passed = datetime.datetime.now() - created_at
    return passed >= delay


class RunResult(Enum):
    NO_CHANGES = 1
    PR_CREATED = 2
    PR_MERGED = 3


class RepoRun:
    def __init__(self, g: git.Git, opts: Options):
        self.git = g
        self.opts = opts

    def execute(self, matcher: task.Task, repo: source.Repository) -> RunResult:
        pr_identifier = repo.find_pull_request(self.git.branch_name)
        if (
            pr_identifier is not None
            and repo.is_pr_closed(pr_identifier) is True
            and matcher.merge_once is True
        ):
            log.info(
                "Existing PR has been closed",
                branch=self.git.branch_name,
                repo=str(repo),
            )
            return RunResult.NO_CHANGES

        if (
            pr_identifier is not None
            and repo.is_pr_merged(pr_identifier) is True
            and matcher.merge_once is True
        ):
            log.info(
                "Existing PR has been merged",
                branch=self.git.branch_name,
                repo=str(repo),
            )
            return RunResult.NO_CHANGES

        work_dir = self.git.prepare(repo)
        tpl_mapping = create_template_mapping(repo)
        apply_actions(repo, matcher, tpl_mapping, work_dir)
        has_changes = self.git.has_changes(work_dir)
        if has_changes is True:
            log.debug("Committing changes", repo=str(repo))
            self.git.commit_changes(work_dir, matcher.commit_msg)
        else:
            log.info("No changes after applying actions", repo=str(repo))

        if (
            self.git.has_changes_base(base_branch=repo.base_branch, repo_dir=work_dir)
            is False
            and pr_identifier is not None
            and repo.is_pr_open(pr_identifier) is True
        ):
            if self.opts.config.dry_run:
                log.warn(
                    "DRY RUN: Closing pull request because base branch contains all changes"
                )
            else:
                log.info(
                    "Closing pull request because base branch contains all changes",
                    repo=str(repo),
                )
                repo.close_pull_request(
                    "Everything up-to-date. Closing.", pr_identifier
                )
                log.info(
                    "Deleting source branch because base branch contains all changes",
                    branch=self.git.branch_name,
                    repo=str(repo),
                )
                repo.delete_branch(pr_identifier)

            return RunResult.NO_CHANGES

        if has_changes is True:
            if self.opts.config.dry_run:
                log.warn("DRY RUN: Not pushing changes")
            else:
                log.debug("Pushing changes", repo=str(repo))
                self.git.push(work_dir)

        pr = source.PullRequest(
            matcher.auto_merge,
            matcher.merge_once,
            matcher.name,
            self.opts.config.pr_title_prefix,
            self.opts.config.pr_title_body.format(matcher_name=matcher.name),
            self.opts.config.pr_title_suffix,
            matcher.pr_body,
            matcher.pr_title,
            auto_merge_after=matcher.auto_merge_after,
        )
        if has_changes is True and (
            pr_identifier is None
            or repo.is_pr_merged(pr_identifier) is True
            or repo.is_pr_closed(pr_identifier) is True
        ):
            if self.opts.config.dry_run:
                log.warn("DRY RUN: Not creating pull request")
            else:
                log.info("Create pull request", repo=str(repo))
                repo.create_pull_request(self.git.branch_name, pr)

            return RunResult.PR_CREATED

        if (
            matcher.auto_merge is True
            and has_changes is False
            and pr_identifier is not None
            and repo.is_pr_open(pr_identifier) is True
        ):
            if not repo.has_successful_pr_build(pr_identifier):
                log.warn(
                    "Cannot merge because build of pull request failed", repo=str(repo)
                )
                return RunResult.NO_CHANGES

            if not can_merge_after(
                repo.pr_created_at(pr_identifier), matcher.auto_merge_after
            ):
                log.info("Too early to merge pull request", repo=str(repo))
                return RunResult.NO_CHANGES

            if not repo.can_merge_pull_request(pr_identifier):
                log.warn("Cannot merge pull request", repo=str(repo))
                return RunResult.NO_CHANGES

            if self.opts.config.dry_run:
                log.warn("DRY RUN: Not merging pull request", repo=str(repo))
            else:
                log.info("Merge pull request", repo=str(repo))
                repo.merge_pull_request(pr_identifier)
                if matcher.delete_branch_after_merge:
                    log.info(
                        "Deleting source branch",
                        branch=self.git.branch_name,
                        repo=str(repo),
                    )
                    repo.delete_branch(pr_identifier)

            return RunResult.PR_MERGED

        if pr_identifier is not None and repo.is_pr_open(pr_identifier) is True:
            repo.update_pull_request(pr_identifier, pr)
            return RunResult.NO_CHANGES

        return RunResult.NO_CHANGES


def apply_actions(
    repo: source.Repository,
    task_: task.Task,
    tpl_mapping: dict,
    work_dir: str,
) -> None:
    for a in task_.actions:
        log.debug(
            "Applying action from task",
            action=a.__class__.__name__,
            task=task_.name,
            repo=str(repo),
        )
        a.apply(work_dir, tpl_mapping)


def execute(opts: Options) -> bool:
    if len(opts.sources) < 1:
        raise RuntimeError(
            "No Source has been configured. Configure access credentials for GitHub or GitLab."
        )

    db = database.new_database(opts.config.database)
    tasks: list[task.Task] = []
    needs_all_repositories = False
    for task_path in opts.task_paths:
        task_ = task.read(task_path)
        task_db = db.get_or_create_task(name=task_.name)
        if task_.enabled is False:
            db.update_task(task_.name, task_.checksum)
            log.info("Task disabled", task=task_.name)
            continue

        if task_.checksum != task_db.checksum:
            needs_all_repositories = True

        tasks.append(task_)

    execution = db.get_last_execution()
    if needs_all_repositories is True or execution.executed_at is None:
        since = datetime.datetime.fromtimestamp(0)
    else:
        since = execution.executed_at

    log.debug("Searching for updated repositories", since=str(since))
    repositories: list[source.Repository] = []
    for s in opts.sources.values():
        known_repos: list[str] = []
        for repository in s.list_repositories(since=since):
            known_repos.append(str(repository))
            repositories.append(repository)

        if needs_all_repositories is False:
            log.debug("Listing repositories with open pull requests")
            for repository in s.list_repositories_with_open_pull_requests():
                if str(repository) in known_repos:
                    continue

                known_repos.append(str(repository))
                repositories.append(repository)

    log.info("Repositories returned by sources", count=len(repositories))
    success = True
    if len(repositories) > 0:
        for task_ in tasks:
            task_success = execute_task(task_, repositories, opts)
            if task_success is True:
                db.update_task(task_.name, task_.checksum)
            else:
                success = False

    if success is False:
        log.error("Errors during execution - check previous log messages")

    ex = database.Execution()
    ex.executed_at = datetime.datetime.utcnow()
    db.save_execution(ex)
    return success


def execute_task(
    task_: task.Task,
    repos: list[source.Repository],
    opts: Options,
) -> bool:
    gitc = git.Git(
        task_.branch(opts.config.git.branch_prefix),
        opts.config.git.clone_options,
        opts.config.git.data_dir,
        opts.config.git.user_name,
        opts.config.git.user_email,
    )
    runner = RepoRun(gitc, opts)
    success = True
    changes_total: int = 0
    for repo in repos:
        try:
            if task_.match(repo) is False:
                log.debug(
                    "Repository does not match", repository=str(repo), task=task_.name
                )
                continue

            log.info("Matched repository", repository=str(repo), task=task_.name)
            result: RunResult = runner.execute(task_, repo)
            if result != RunResult.NO_CHANGES:
                changes_total += 1

            if task_.change_limit is not None and changes_total >= task_.change_limit:
                log.info(
                    "Limit of changes reached",
                    limit=task_.change_limit,
                    task=task_.name,
                )
                return success

        except Exception:
            log.exception("Task failed", repository=str(repo), task=task_.name)
            success = False

    return success


def options_from_config(path: str) -> Options:
    cfg = config.read_config_from_file(path)
    return config_to_options(cfg)


def config_to_options(cfg: config.Config) -> Options:
    opts = Options(cfg)

    opts.encoding_registry = encoding.Registry()
    opts.encoding_registry.register(
        encoding.Json(cfg.json_.indent), cfg.json_.extensions
    )
    opts.encoding_registry.register(encoding.Toml(), cfg.toml.extensions)
    opts.encoding_registry.register(
        encoding.Yaml(cfg.yaml.explicit_start), cfg.yaml.extensions
    )

    if cfg.github.access_token != "":
        source_github = source.Github(cfg.github.access_token, cfg.github.base_url)
        opts.sources["github"] = source_github

    if cfg.gitlab.private_token != "":
        source_gitlab = source.Gitlab(cfg.gitlab.url, cfg.gitlab.private_token)
        opts.sources["gitlab"] = source_gitlab

    return opts


def create_template_mapping(repo: source.Repository) -> dict[str, str]:
    return {
        "repo_name": repo.name,
        "repo_project": repo.project,
        "repo_source": repo.source,
    }



from rcmt import task


def validate(task_file_paths: tuple[str]) -> bool:
    result: bool = False
    for task_file_path in task_file_paths:
        try:
            t = task.read(task_file_path)
            print(f"✅ Task {t.name} in file {task_file_path} is valid")
        except Exception as e:
            print(f"❌ Task in file {task_file_path} is invalid: {str(e)}")
            result = False

    return result



import datetime
import hashlib
import importlib.machinery
import importlib.util
import os.path
import random
import string
import sys
from typing import Optional

from slugify import slugify

from rcmt import action, matcher, source
from rcmt.fs import FileProxy


class Task:
    """
    A Task connects Actions with repositories. rcmt reads the Task, finds matching
    repositories and then applies its Actions to each repository.

    :param name: The name of the Task. rcmt uses the name to identify a task.
    :param auto_merge: rcmt automatically merges a pull request on its next run. The
                       pull request must pass all its checks.
    :param auto_merge_after: A duration after which to automatically merge a Pull
                             Request. Requires ``auto_merge`` to be set to ``true``.
    :param branch_name: Name of the branch in git. Defaults to ``branch_prefix`` +
                        ``name``.
    :param change_limit: Limits the number of changes per run of the Task. A change is
                         either a pull request created or merged. Helps to reduce the
                         load on a CI/CD system when a Task would create a lot of pull
                         requests at the same time. Defaults to ``None`` which means no
                         limit.
    :param commit_msg: Message to use when committing changes via git.
    :param delete_branch_after_merge: If ``True``, rcmt will delete the branch after it
                                      has been merged. Defaults to ``True``.
    :param enabled: If ``False``, disables the task. Handy if a task needs to be stopped
                    temporarily. Defaults to ``True``.
    :param merge_once: If ``True``, rcmt does not create another pull request if it
                       created a pull request for the same branch before and that pull
                       request has been merged.
    :param pr_body: Define a custom body of a pull request.
    :param pr_title: Set a custom title for a pull request.

    **Example**

    .. code-block:: python

       from datetime import timedelta

       from rcmt import Task
       from rcmt.matcher import FileExists, RepoName

       with Task(
           name="python-defaults",
           auto_merge=True,
           auto_merge_after=timedelta(days=7)
       ) as task:
           task.add_matcher(FileExists("pyproject.toml"))
           task.add_matcher(RepoName("^github.com/wndhydrnt/rcmt$"))

           task.pr_title = "A custom PR title"
           task.pr_body = '''A custom PR title.
           It supports multiline strings.'''
    """

    def __init__(
        self,
        name: str,
        auto_merge: bool = False,
        auto_merge_after: Optional[datetime.timedelta] = None,
        branch_name: str = "",
        change_limit: Optional[int] = None,
        commit_msg: str = "Applied actions",
        delete_branch_after_merge: bool = True,
        enabled: bool = True,
        merge_once: bool = False,
        pr_body: str = "",
        pr_title: str = "",
    ):
        self.auto_merge = auto_merge
        self.auto_merge_after = auto_merge_after
        self.branch_name = branch_name
        self.change_limit = change_limit
        self.commit_msg = commit_msg
        self.delete_branch_after_merge = delete_branch_after_merge
        self.enabled = enabled
        self.merge_once = merge_once
        self.name = name
        self.pr_body = pr_body
        self.pr_title = pr_title

        self.actions: list[action.Action] = []
        self.checksum: str = ""
        self.file_proxies: list[FileProxy] = []
        self.matchers: list[matcher.Base] = []

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass

    def add_action(self, a: action.Action) -> None:
        """
        Add an Action to apply to every matching repository.

        :param a: The Action.
        """
        self.actions.append(a)

    def add_matcher(self, m: matcher.Base) -> None:
        """
        Add a Matcher that matches repositories.

        :param m: The matcher to add.
        """
        self.matchers.append(m)

    def branch(self, prefix: str) -> str:
        if self.branch_name != "":
            return self.branch_name

        return f"{prefix}{slugify(self.name)}"

    def load_file(self, path: str) -> FileProxy:
        """
        Returns a proxy that an Action can use to load a file.

        :param path: Path to the file to load. Relative to the file that contains the
                     Task.
        """
        fp = FileProxy(path)
        self.file_proxies.append(fp)
        return fp

    def match(self, repo: source.Repository) -> bool:
        for m in self.matchers:
            if m.match(repo) is False:
                return False

        return True

    def set_path(self, path):
        """
        Set the path to the Task.
        Forwards this path to all ``FileProxys`` created when calling the
        ``load_file`` function.
        rcmt calls this function when it loads a Task file.

        :param path: Path to the directory that contains the Task file.
        """
        for fp in self.file_proxies:
            fp.set_path(path)


class Run(Task):
    """
    Run is an alias of Task.

    Provides backwards compatibility with task files written for rcmt <= 0.15.3.
    """

    pass


def read(path: str) -> Task:
    checksum = hashlib.md5()
    with open(path) as f:
        while True:
            line = f.readline()
            if line == "":
                break

            checksum.update(line.encode("utf-8"))

    rndm = "".join(random.choice(string.ascii_lowercase) for _ in range(8))
    mod_name = f"rcmt_task_{rndm}"
    loader = importlib.machinery.SourceFileLoader(mod_name, path)
    spec = importlib.util.spec_from_loader(mod_name, loader)
    assert spec is not None
    new_module = importlib.util.module_from_spec(spec)
    sys.modules[spec.name] = new_module
    loader.exec_module(new_module)
    try:
        task = new_module.task  # type: ignore # because the content of module is not known
    except AttributeError:
        try:
            # Accept variable "run" for backward compatibility with versions <= 0.15.3
            task = new_module.run  # type: ignore # because the content of module is not known
        except AttributeError:
            raise RuntimeError(f"Task file {path} does not define variable 'task'")

    if not isinstance(task, Task):
        raise RuntimeError(
            f"Task file {path} defines variable 'task' but is not of type Task"
        )

    task.set_path(os.path.dirname(path))
    task.checksum = checksum.hexdigest()
    return task  # type: ignore # because the content of module is not known



"""Add model "Extension"

Revision ID: b3117b16eb76
Revises:
Create Date: 2022-12-05 20:37:00.859534

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "b3117b16eb76"
down_revision = None
branch_labels = None
depends_on = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "executions",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("executed_at", sa.DateTime(), nullable=False),
        sa.PrimaryKeyConstraint("id"),
    )
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table("executions")
    # ### end Alembic commands ###



import sys

import click
import structlog

import rcmt
from rcmt.log import configure as configure_logging

# Default logging settings before any configuration has been read.
configure_logging(format=None, level="error")
log: structlog.stdlib.BoundLogger = structlog.get_logger()


run_help = """Apply a Task to all matching repositories of a remote Git host.

rcmt will find all matching repositories, clone them, apply each Action from
the Task file and create a pull request if files have changed.

Examples:

\b
# Apply Task "task.py"
rcmt run --config ./config.yaml ./task.py
"""


@click.command(
    help=run_help,
    short_help="Apply a Task to all matching repositories of a remote Git host.",
)
@click.option("--config", help="Path to configuration file.", default="", type=str)
@click.argument("task_file", nargs=-1)
def run(config: str, task_file: list[str]):
    try:
        opts = rcmt.options_from_config(config)
        opts.task_paths = task_file
        configure_logging(format=opts.config.log_format, level=opts.config.log_level)
        result = rcmt.execute(opts)
        if result is False:
            exit(1)
    except Exception:
        log.exception("Unexpected error")
        exit(1)


validate_help = """Validate Task files.

Test that rcmt can load a task file. Useful during CI/CD before rolling out a change to
production.

No extra configuration is necessary to execute this command. A Task is loaded but no
calls to external APIs are made.

Note: Code in module scope of a Task will be executed during validation. A Task file
with the following content will print "Hello" when validated:

\b
print("Hello")
with Task(name="example") as task:
    # Add matchers/actions here



Examples:

\b
# Validate a single Task file
rcmt validate task.py

\b
# Validate multiple Task files
rcmt validate task1.py task2.py

\b
# Validate all Task files in a directory
rcmt validate tasks/*.py
"""


@click.command(help=validate_help, short_help="Validate Task files")
@click.argument("task_file", type=click.Path(), nargs=-1)
def validate(task_file: tuple[str]):
    if rcmt.validate(task_file_paths=task_file) is False:
        exit(1)


verify_help = """Verify a Task locally.

This command makes it possible to quickly verify that Matchers and Actions of a Task
work for a given repository.

The command queries the API of a Source and clones the repository, but does not create a
pull request.

Each execution of "verify" restores the cloned repository to a clean state before
applying Actions.

Examples:

\b
# Verify that Matchers and Actions in task.py work for the repository "github.com/wndhydrnt/rcmt"
rcmt verify gitlab.com/wandhydrant/rcmt-test task.py

Note: rcmt needs to query the API of a Source, like GitHub or GitLab, to execute the
Task.

\b
# Supply a GitHub Access Token
RCMT_GITHUB__ACCESS_TOKEN=xxx rcmt verify github.com/wndhydrnt/rcmt task.py

\b
# Supply a configuration file that contains an access token
rcmt verify --config config.yaml gitlab.com/wandhydrant/rcmt-test task.py
"""


@click.command(
    help=verify_help,
    short_help="Verify a Task locally. Useful during development.",
)
@click.option("--config", help="Path to configuration file.", default="", type=str)
@click.option(
    "--directory",
    help="Directory that stores cloned git repositories.",
    default=".rcmt",
    type=str,
)
@click.argument("repository", type=str)
@click.argument("task_file", type=str)
def verify(config: str, directory: str, repository: str, task_file: str):
    try:
        opts = rcmt.options_from_config(config)
        opts.task_paths = [task_file]
        configure_logging(format=opts.config.log_format, level=opts.config.log_level)
        rcmt.execute_verify(
            directory=directory, opts=opts, out=sys.stdout, repo_name=repository
        )
    except Exception:
        log.exception("Unexpected error")
        exit(1)


@click.command()
def version():
    click.echo(rcmt.__version__)


@click.group()
def main():
    pass


main.add_command(run)
main.add_command(validate)
main.add_command(verify)
main.add_command(version)



from typing import Optional, TextIO

import structlog

import rcmt.git
from rcmt import task
from rcmt.rcmt import Options, apply_actions, create_template_mapping
from rcmt.source import Repository

log: structlog.stdlib.BoundLogger = structlog.get_logger()


def execute(directory: str, opts: Options, out: TextIO, repo_name: str) -> None:
    repository: Optional[Repository] = None
    for source_name, source in opts.sources.items():
        repository = source.create_from_name(repo_name)
        if repository is not None:
            break

        log.debug("Source did not return a repository", source=source_name)

    if repository is None:
        log.error(f"No Source found for repository {repo_name}")
        return

    for task_path in opts.task_paths:
        result: bool = True
        t = task.read(task_path)
        for m in t.matchers:
            if m.match(repository) is True:
                print(f"✅ Matcher {str(m)} matches", file=out)
            else:
                result = False
                print(f"❌ Matcher {str(m)} does not match", file=out)

        if result is False:
            print(
                f"❌ - at least one Matcher did not match repository {repo_name}",
                file=out,
            )
            return

        print("---")
        gitc = rcmt.git.Git(
            t.branch(opts.config.git.branch_prefix),
            opts.config.git.clone_options,
            directory,
            opts.config.git.user_name,
            opts.config.git.user_email,
        )

        print("🏗️  Preparing git clone", file=out)
        checkout_dir = gitc.prepare(repository)
        tpl_mapping: dict[str, str] = create_template_mapping(repository)
        print("🚜 Applying actions", file=out)
        apply_actions(
            repo=repository, task_=t, tpl_mapping=tpl_mapping, work_dir=checkout_dir
        )
        if gitc.has_changes(repo_dir=checkout_dir):
            print(
                f"😍 Actions modified files - view changes in {checkout_dir}", file=out
            )
        else:
            print("🤔 No changes after applying Actions", file=out)



import datetime
import urllib.parse
from typing import Any, Generator, Optional, TextIO, Union

import humanize


class PullRequest:
    def __init__(
        self,
        auto_merge: bool,
        merge_once: bool,
        run_name: str,
        title_prefix: str,
        title_body: str,
        title_suffix: str,
        custom_body: str = "",
        custom_title: str = "",
        auto_merge_after: Optional[datetime.timedelta] = None,
    ):
        self.auto_merge = auto_merge
        self.auto_merge_after = auto_merge_after
        self.custom_body = custom_body
        self.custom_title = custom_title
        self.merge_once = merge_once
        self.run_name = run_name
        self.title_prefix = title_prefix
        self.title_body = title_body
        self.title_suffix = title_suffix

    def __eq__(self, other: object) -> bool:
        """
        Aids with checking equality during unit tests.

        :param other: Other PullRequest
        :return: bool
        """
        if self.auto_merge != getattr(other, "auto_merge"):
            return False

        if self.auto_merge_after != getattr(other, "auto_merge_after"):
            return False

        if self.custom_body != getattr(other, "custom_body"):
            return False

        if self.custom_title != getattr(other, "custom_title"):
            return False

        if self.merge_once != getattr(other, "merge_once"):
            return False

        if self.run_name != getattr(other, "run_name"):
            return False

        if self.title_prefix != getattr(other, "title_prefix"):
            return False

        if self.title_body != getattr(other, "title_body"):
            return False

        if self.title_suffix != getattr(other, "title_suffix"):
            return False

        return True

    @property
    def body(self) -> str:
        if self.custom_body == "":
            body = f"Apply changes from Run {self.run_name}\n"
        else:
            body = self.custom_body
            body += "\n"

        body += "\n---\n\n"

        body += "**Automerge:** "
        # Add two spaces at the end of each line to make GitLab create a line break.
        if self.auto_merge is True:
            if self.auto_merge_after is None:
                body += "Enabled. rcmt merges this automatically on its next run and if all checks have passed.  \n"
            else:
                after = humanize.naturaldelta(self.auto_merge_after)
                body += f"Enabled. rcmt automatically merges this in {after} and if all checks have passed.  \n"
        else:
            body += "Disabled. Merge this manually.  \n"

        if self.merge_once is True:
            body += "**Ignore:** Close this PR and it will not be recreated again.  \n"
        else:
            body += "**Ignore:** This PR will be recreated if closed.  \n"

        body += """
---

_This pull request has been created by [rcmt](https://rcmt.readthedocs.io/)._"""

        return body

    @property
    def title(self) -> str:
        if self.custom_title != "":
            return self.custom_title

        return f"{self.title_prefix} {self.title_body} {self.title_suffix}".strip()


class Repository:
    """
    Repository provides all methods needed to interact with a single repository of a
    Source.
    """

    def __str__(self):
        return self.full_name

    @property
    def base_branch(self) -> str:
        """
        :return: Name of the base branch of this repository.
        :rtype: str
        """
        raise NotImplementedError("class does not implement Repository.base_branch()")

    def can_merge_pull_request(self, identifier: Any) -> bool:
        """
        Checks if a pull request can be merged.

        Use this function to check if a Source, e.g. GitHub, indicates that merge
        conflicts exist.

        :param identifier: Data to identify the pull request as returned by find_pull_request.
        :return: Indicates if the pull request can be merged.
        :rtype: bool
        """
        raise NotImplementedError(
            "class does not implement Repository.can_merge_pull_request()"
        )

    @property
    def clone_url(self) -> str:
        """
        :return: Url to clone this repository.
        :rtype: str
        """
        raise NotImplementedError("class does not implement Repository.clone_url()")

    def close_pull_request(self, message: str, pr: Any) -> None:
        """
        Closes a pull request and adds a comment that explains why the pull request has
        been closed.

        :param message: Message to post as a comment.
        :param pr: The pull request.
        :rtype: None
        """
        raise NotImplementedError(
            "class does not implement Repository.close_pull_request()"
        )

    def create_pull_request(self, branch: str, pr: PullRequest) -> None:
        """
        Creates a pull request for the given branch.

        :param branch: Name of the branch.
        :param pr: The pull request.
        :rtype: None
        """
        raise NotImplementedError(
            "class does not implement Repository.create_pull_request()"
        )

    def delete_branch(self, identifier: Any) -> None:
        raise NotImplementedError(
            "class does not implement Repository.delete_pull_request()"
        )

    def find_pull_request(self, branch: str) -> Union[Any, None]:
        """
        Finds and returns the pull request opened by rcmt.

        :param branch: Name of the branch from which a pull request has been created.
        :return: Implementation of this method should return None if no pull request is open. Any other value will be
                 passed to merge_pull_request() so it can identify which PR to merge.
        :rtype: Any, None
        """
        raise NotImplementedError(
            "class does not implement Repository.has_open_pull_request()"
        )

    @property
    def full_name(self) -> str:
        """
        :return: Full name of the repository.
        :rtype: str
        """
        return f"{self.source}/{self.project}/{self.name}"

    def get_file(self, path: str) -> TextIO:
        raise NotImplementedError("class does not implement Repository.has_file()")

    def has_file(self, path: str) -> bool:
        """
        Checks if a file exists in a repository.

        rcmt calls this function when matching repositories. This is more efficient than
        checking out the whole repository to check if a file exists.

        :param path: Path to a file or directory in the repository.
        :return: Indicates if the files exists.
        :rtype: bool
        """
        raise NotImplementedError("class does not implement Repository.has_file()")

    def has_successful_pr_build(self, identifier: Any) -> bool:
        """
        Checks if a pull request has passed all checks. rcmt will call merge_pull_request if this function returns
        True.

        :param identifier: Data to identify the pull request as returned by find_pull_request.
        :return: Indicates if the build is successful.
        :rtype: bool
        """
        raise NotImplementedError(
            "class does not implement Repository.has_successful_build()"
        )

    def is_pr_closed(self, identifier: Any) -> bool:
        """
        Checks if a pull request has been closed by the user.
        A pull request is closed if a user manually closes it without merging it.
        If a user has closed a pull request without merging it, rcmt will not create a
        new one.

        :param identifier: Data to identify the pull request as returned by find_pull_request.
        :return: Indicates that a pull request has been closed by the user.
        :rtype: bool
        """
        raise NotImplementedError("class does not implement Repository.is_pr_closed()")

    def is_pr_merged(self, identifier: Any) -> bool:
        """
        Checks if a pull request has been merged.
        rcmt uses this information to determine if it can create new pull requests.

        :param identifier: Data to identify the pull request as returned by find_pull_request.
        :return: Indicates that a pull request has been merged.
        :rtype: bool
        """
        raise NotImplementedError("class does not implement Repository.is_pr_merged()")

    def is_pr_open(self, identifier: Any) -> bool:
        """
        Checks if a pull request is open.
        rcmt will attempt to merge the pull request if this method returns true and
        there are no new changes from Actions.

        :param identifier: Data to identify the pull request as returned by find_pull_request.
        :return: Indicates that a pull request is open.
        :rtype: bool
        """
        raise NotImplementedError("class does not implement Repository.is_pr_open()")

    def merge_pull_request(self, identifier: Any) -> None:
        """
        Merges a pull request.

        :param identifier: Data to identify the pull request as returned by find_pull_request.
        """
        raise NotImplementedError(
            "class does not implement Repository.merge_pull_request()"
        )

    @property
    def name(self) -> str:
        """
        :return: Name of the repository.
        :rtype: str
        """
        raise NotImplementedError("class does not implement Repository.name()")

    def pr_created_at(self, pr: Any) -> datetime.datetime:
        """
        Returns the date and time at which the pull request was created.

        :param pr: The pull request identifier as returned by find_pull_request().
        :return: Date and time at which the pull request was created.
        :rtype: datetime.datetime
        """
        raise NotImplementedError("class does not implement Repository.pr_created_at()")

    @property
    def project(self) -> str:
        """
        :return: Name of the project the repository belongs to.
        :rtype: str
        """

        raise NotImplementedError("class does not implement Repository.project()")

    @property
    def source(self) -> str:
        """
        :return: Name of the source that hosts the repository.
        :rtype: str
        """
        raise NotImplementedError("class does not implement Repository.source()")

    def update_pull_request(self, pr: Any, pr_data: PullRequest) -> None:
        """
        Updates the title and body of a pull request.

        To reduce calls to the API, the implementation of this function should check if
        title or body have changed and only update the pull request if necessary.

        :param pr: Data to identify the pull request as returned by find_pull_request.
        :param pr_data: Data of the pull request that rcmt wants to create/update.
        :rtype: None
        """
        raise NotImplementedError(
            "class does not implement Repository.update_pull_request()"
        )


class Base:
    """
    Base defines the methods every Source needs to implement.

    """

    def create_from_name(self, name: str) -> Optional[Repository]:
        """
        Crate an instance of Repository from the name of a repository.

        :param name: Name of the repository.
        :return: An instance of Repository.
        :rtype: rcmt.source.Repository
        """
        raise NotImplementedError("class does not implement Base.create_from_name()")

    def list_repositories_with_open_pull_requests(
        self,
    ) -> Generator[Repository, None, None]:
        raise NotImplementedError(
            "class does not implement Base.list_open_pull_requests()"
        )

    def list_repositories(self, since: datetime.datetime) -> list[Repository]:
        """
        :param since: Date and time of the last run of rcmt to find only those repositories that have received and update since.

        :return: List of all known repositories.
        :rtype: list[rcmt.source.Repository]
        """
        raise NotImplementedError("class does not implement Base.list_repositories()")


def add_credentials_to_url(url: str, password: str, username: str = "") -> str:
    parsed = urllib.parse.urlparse(url)
    if username == "":
        credentials = password
    else:
        credentials = f"{username}:{password}"

    return parsed._replace(netloc=f"{credentials}@{parsed.netloc}").geturl()



import datetime
import fnmatch
import io
from typing import Any, Generator, Optional, TextIO, Union

import github
import github.PullRequest
import github.Repository
import structlog

from ..log import SECRET_MASKER
from .source import Base, PullRequest, Repository, add_credentials_to_url

log = structlog.get_logger(source="github")


class GithubRepository(Repository):
    def __init__(self, access_token: str, repo: github.Repository.Repository):
        self.access_token = access_token
        self.repo = repo
        self._name = repo.name
        self._project = repo.owner.login

    @property
    def base_branch(self) -> str:
        return self.repo.default_branch

    def can_merge_pull_request(
        self, identifier: github.PullRequest.PullRequest
    ) -> bool:
        if identifier.mergeable is None:
            return True

        if identifier.mergeable is False:
            log.warn(
                "GitHub indicates that the PR is not mergeable",
                pr_id=identifier.id,
                repo=str(self),
            )

        return identifier.mergeable

    @property
    def clone_url(self):
        return add_credentials_to_url(
            url=self.repo.clone_url, password=self.access_token
        )

    def close_pull_request(
        self, message: str, pr: github.PullRequest.PullRequest
    ) -> None:
        pr.create_issue_comment(body=message)
        pr.edit(state="closed")

    def create_pull_request(self, branch: str, pr: PullRequest):
        log.debug(
            "Creating pull request", base=self.base_branch, head=branch, repo=str(self)
        )
        self.repo.create_pull(
            title=pr.title,
            body=pr.body,
            base=self.base_branch,
            head=branch,
            maintainer_can_modify=True,
        )

    def delete_branch(self, identifier: github.PullRequest.PullRequest) -> None:
        if self.repo.delete_branch_on_merge is False:
            self.repo.get_git_ref(ref=f"heads/{identifier.head.ref}").delete()

    def find_pull_request(self, branch: str) -> Union[Any, None]:
        log.debug("Listing pull requests", repo=str(self))
        for pr in self.repo.get_pulls(state="all"):
            if pr.head.ref == branch:
                return pr

        return None

    def get_file(self, path: str) -> TextIO:
        try:
            file = self.repo.get_contents(path=path, ref=self.base_branch)
        except github.UnknownObjectException:
            raise FileNotFoundError("file does not exist in repository")

        if isinstance(file, list):
            if len(file) != 1:
                raise FileNotFoundError("github returned more than one file")

            file = file[0]

        return io.StringIO(file.decoded_content.decode("utf-8"))

    def has_file(self, path: str) -> bool:
        try:
            tree = self.repo.get_git_tree(self.base_branch, True)
            for entry in tree.tree:
                if fnmatch.fnmatch(entry.path, path):
                    return True
        except github.GithubException as e:
            if e.status == 409:
                log.warning("Tree not found - empty repository?", repo=str(self))
                return False
            else:
                raise e

        return False

    def has_successful_pr_build(self, pr: github.PullRequest.PullRequest) -> bool:
        log.debug("Checking PR builds", repo=str(self))
        for cr in self.repo.get_commit(pr.head.sha).get_check_runs():
            if cr.conclusion != "success":
                log.debug(
                    "GitHub check not successful",
                    repo=str(self),
                    check_name=cr.name,
                    check_conclusion=cr.conclusion,
                )
                return False

        return True

    def is_pr_closed(self, pr: github.PullRequest.PullRequest) -> bool:
        return pr.state == "closed" and pr.merged is False

    def is_pr_merged(self, pr: github.PullRequest.PullRequest) -> bool:
        return pr.state == "closed" and pr.merged is True

    def is_pr_open(self, pr: github.PullRequest.PullRequest) -> bool:
        return pr.state == "open"

    def merge_pull_request(self, pr: github.PullRequest.PullRequest) -> None:
        log.debug("Merging pull request", repo=str(self))
        pr.merge(commit_title="Auto-merge by rcmt")

    @property
    def name(self) -> str:
        return self.repo.name

    def pr_created_at(self, pr: github.PullRequest.PullRequest) -> datetime.datetime:
        return pr.created_at

    @property
    def project(self) -> str:
        return self.repo.owner.login

    @property
    def source(self) -> str:
        return "github.com"

    def update_pull_request(
        self, pr: github.PullRequest.PullRequest, pr_data: PullRequest
    ) -> None:
        needs_update = False
        if pr.title != pr_data.title:
            needs_update = True

        if pr.body != pr_data.body:
            needs_update = True

        if needs_update is True:
            log.debug("Updating PR data", pr_id=pr.id, repo=str(self))
            pr.edit(title=pr_data.title, body=pr_data.body)


class Github(Base):
    def __init__(self, access_token: str, base_url: str):
        self.access_token = access_token
        self.client = github.Github(login_or_token=access_token, base_url=base_url)

        SECRET_MASKER.add_secret(access_token)

    def create_from_name(self, name: str) -> Optional[Repository]:
        repo_name = "/".join(name.split("/")[1:])
        try:
            gh_repo = self.client.get_repo(full_name_or_id=repo_name, lazy=False)
        except github.UnknownObjectException:
            return None

        return GithubRepository(access_token=self.access_token, repo=gh_repo)

    def list_repositories_with_open_pull_requests(
        self,
    ) -> Generator[Repository, None, None]:
        user = self.client.get_user()
        for issue in self.client.search_issues(
            f"is:open is:pr author:{user.login} archived:false"
        ):
            yield GithubRepository(self.access_token, issue.repository)

    def list_repositories(self, since: datetime.datetime) -> list[Repository]:
        log.debug("start fetching repositories")
        repos: list[Repository] = []
        for gh_repo in self.client.get_user().get_repos(
            direction="desc", sort="updated"
        ):
            if gh_repo.updated_at > since:
                repos.append(GithubRepository(self.access_token, gh_repo))
            else:
                break

        log.debug("finished fetching repositories")
        return repos



import logging
import sys
from typing import Any, Mapping, MutableMapping, Optional

import structlog


class SecretMasker:
    def __init__(self):
        self.secrets: list[str] = []

    def add_secret(self, s: str) -> None:
        self.secrets.append(s)

    def process_event(
        self, _, __, event_dict: MutableMapping[str, Any]
    ) -> Mapping[str, Any]:
        for k, v in event_dict.items():
            if not isinstance(v, str):
                continue

            for secret in self.secrets:
                if secret in v:
                    event_dict[k] = v.replace(secret, "****")

        return event_dict


SECRET_MASKER = SecretMasker()


def configure(format: Optional[str], level: str) -> None:
    processors = [
        SECRET_MASKER.process_event,
        structlog.processors.add_log_level,
        # Add a timestamp in ISO 8601 format.
        structlog.processors.TimeStamper(fmt="iso"),
        # If the "stack_info" key in the event dict is true, remove it and
        # render the current stack trace in the "stack" key.
        structlog.processors.StackInfoRenderer(),
        # If the "exc_info" key in the event dict is either true or a
        # sys.exc_info() tuple, remove "exc_info" and render the exception
        # with traceback into the "exception" key.
        structlog.processors.format_exc_info,
        # If some value is in bytes, decode it to a unicode str.
        structlog.processors.UnicodeDecoder(),
    ]

    use_json = False
    if format is None:
        if sys.stderr.isatty() is False:
            use_json = True
    else:
        if format.lower() == "json":
            use_json = True

    if use_json is True:
        processors.append(structlog.processors.EventRenamer("message"))
        processors.append(structlog.processors.JSONRenderer())
    else:
        processors.append(
            structlog.dev.ConsoleRenderer(
                colors=sys.stderr is not None and sys.stderr.isatty()
            )
        )

    log_level = logging.getLevelName(level.upper())
    structlog.configure(
        processors=processors,  # type: ignore
        # `wrapper_class` is the bound logger that you get back from
        # get_logger(). This one imitates the API of `logging.Logger`.
        wrapper_class=structlog.make_filtering_bound_logger(log_level),
        # `logger_factory` is used to create wrapped loggers that are used for
        # OUTPUT. This one returns a `logging.Logger`. The final value (a JSON
        # string) from the final processor (`JSONRenderer`) will be passed to
        # the method of the same name as that you've called on the bound logger.
        logger_factory=structlog.PrintLoggerFactory(sys.stderr),
        # Effectively freeze configuration after creating the first bound
        # logger.
        cache_logger_on_first_use=True,
    )
